#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Base url to retrieve data (DGC)
BASE_URL_DGC = "https://get.dgc.gov.it/v1/dgc/"
# Alternative base url to retrieve data (DGC)
BASE_URL_DGCG = "https://dgcg.covidbevis.se/tp/"
# Base url to retrieve data (NHS)
BASE_URL_NHS = "https://covid-status.service.nhsx.nhs.uk/"



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import sys
import zlib
import base45
import json
import cbor2
import pytz

from datetime import datetime
from cose.headers import KID, Algorithm
from cose.messages import CoseMessage

from greenpass.data import TestType
from greenpass.data import GreenPassKeyManager


class UnrecognizedException(Exception):
    def __init__(self, m):
        """The greenpass was unrecognized."""
        super(UnrecognizedException, self).__init__(m)


class TestResult(object):
    def __init__(self, t):
        """Get the test result parsing the certificate data."""
        self.t = int(t)

    def is_positive(self):
        return self.t == 260373001

    def is_negative(self):
        return self.t == 260415000

    def is_aladeen(self):
        return not self.is_positive() and not self.is_negative()

    def is_unknown(self):
        return self.is_aladeen()

    def __str__(self):
        """Readable format for test result."""
        if self.is_positive():
            return "Positive"
        elif self.is_negative():
            return "Negative"
        return "Unknown"


def _parse_timestamp(d):
    return datetime.fromtimestamp(d, pytz.utc)


def _parse_date_time(d):
    try:
        # strptime on Python < 3.7 does not correctly parse the
        # timezone format with a colon inside, remove the colon
        # in the timezone specification.
        rgx = r"""
        (
            \d{4}-\d{2}-\d{2}T # Date
            \d{2}:\d{2}:\d{2}  # Time
        )
        (
            [+-]
            \d{2}              # Timezone hours
            (?::\d{2})         # Timezone minutes
        )?
        """
        compiled_regex = re.compile(rgx, re.VERBOSE)
        r = compiled_regex.match(d)
        date = r.group(1)
        zone = r.group(2)

        if zone is None:
            date += "+0000"
        else:
            date += r.group(2).replace(":", "")

        testcollectiondate = datetime.strptime(
            date, "%Y-%m-%dT%H:%M:%S%z"
        )
    except Exception as e:
        print(e, file=sys.stderr)
        testcollectiondate = 0
    return testcollectiondate


def _parse_date(d):
    if "T" in d:
        d = _parse_date_time(d)
    else:
        d = datetime.strptime(d, "%Y-%m-%d")
        d = pytz.utc.localize(d, is_dst=None).astimezone(pytz.utc)
    return d


class Certificate(object):
    def __init__(self, k):
        """Certificate class transcription."""
        self.k = k
        self.release_date = None
        self.expiration_date = None
        self.dose_number = None
        self.total_doses = None
        self.validity_from = None
        self.validity_until = None
        self.vaccine_pn = None
        self.vaccine_date = None
        self.test_type = None
        self.target_disease = None
        self.test_collection_date = None
        self.certificate_id = None
        self.expired = False
        self.hours_to_valid = None
        self.remaining_hours = None
        self.blocklisted = False
        self.sign_alg = None
        self.kid = None
        self._type = None
        self.test_result = None
        self.verified = None
        self.manufacturer = None
        self.date_of_collection = None
        self.vaccination_date = None
        self.vaccine_type = None

        self.info = {
            "qr": {},
            "personal": {},
            "cert": {}
        }

        # Filter table to process data
        self.filter_table = {
            self.k.get_target_disease()[0]:     self.set_target_disease,
            self.k.get_dose_number()[0]:        self.set_dose_number,
            self.k.get_total_doses()[0]:        self.set_total_doses,
            self.k.get_certificate_id()[0]:     self.set_certificate_id,
            self.k.get_validity_from()[0]:      self.set_validity_from,
            self.k.get_validity_until()[0]:     self.set_validity_until,
            self.k.get_vaccine_type()[0]:       self.set_vaccine_type,
            self.k.get_test_type()[0]:          self.set_test_type,
            self.k.get_manufacturer()[0]:       self.set_manufacturer,
            self.k.get_vaccine_pn()[0]:         self.set_vaccine_pn,
            self.k.get_release_date()[0]:       self.set_release_date,
            self.k.get_expiration_date()[0]:    self.set_expiration_date,
            self.k.get_date_of_collection()[0]: self.set_date_of_collection,
            self.k.get_vaccination_date()[0]:   self.set_vaccination_date,
            self.k.get_test_result()[0]:        self.set_test_result,
        }

    # Unstructured data (just for print, not used)
    def set_info(self, _type, key, val):
        self.info[_type][key] = val

    def get_info(self):
        return self.info

    def get_target_disease(self):
        return self.target_disease

    def set_target_disease(self, val):
        self.target_disease = val

    def get_release_date(self):
        return self.release_date

    def set_release_date(self, val):
        self.release_date = _parse_timestamp(val)

    def get_expiration_date(self):
        return self.expiration_date

    def set_expiration_date(self, val):
        self.expiration_date = _parse_timestamp(val)

    def get_dose_number(self):
        return self.dose_number

    def set_dose_number(self, val):
        self.dose_number = val

    def get_total_doses(self):
        return self.total_doses

    def set_total_doses(self, val):
        self.total_doses = val

    def get_validity_from(self):
        return self.validity_from

    def set_validity_from(self, val):
        self.validity_from = _parse_date(val)

    def get_validity_until(self):
        return self.validity_until

    def set_validity_until(self, val):
        self.validity_until = _parse_date(val)

    def get_vaccine_pn(self):
        return self.vaccine_pn

    def set_vaccine_pn(self, val):
        self.vaccine_pn = val

    def get_vaccination_date(self):
        return self.vaccination_date

    def set_vaccination_date(self, val):
        self.vaccination_date = _parse_date(val)

    def get_test_type(self):
        return self.test_type

    def set_test_type(self, val):
        self.test_type = val

    def get_manufacturer(self):
        return self.manufacturer

    def set_manufacturer(self, val):
        self.manufacturer = val

    def get_vaccine_type(self):
        return self.vaccine_type

    def set_vaccine_type(self, val):
        self.vaccine_type = val

    def get_test_result(self):
        return self.test_result

    def set_test_result(self, val):
        self.test_result = TestResult(val)

    def get_date_of_collection(self):
        return self.date_of_collection

    def set_date_of_collection(self, val):
        self.date_of_collection = _parse_date(val)

    def get_certificate_id(self):
        return self.certificate_id

    def set_certificate_id(self, val):
        self.certificate_id = val

    def get_expired(self):
        return self.expired

    def set_expired(self, val):
        self.expired = val

    def get_hours_to_valid(self):
        return self.hours_to_valid

    def set_hours_to_valid(self, val):
        self.hours_to_valid = val

    def get_remaining_hours(self):
        return self.remaining_hours

    def set_remaining_hours(self, val):
        self.remaining_hours = val

    def get_blocklisted(self):
        return self.blocklisted

    def set_blocklisted(self, val):
        self.blocklisted = val

    def get_sign_alg(self):
        return self.sign_alg

    def set_sign_alg(self, val):
        self.sign_alg = val

    def get_kid(self):
        return self.kid

    def set_kid(self, val):
        self.kid = val

    def get_parent(self):
        return self.parent

    def set_parent(self, p):
        self.parent = p

    def set_key(self, k):
        self.parent.set_key(k)

    def verify(self):
        return self.parent.verify()

    def get_type(self):
        # Calculate type if not set
        if self._type is None:
            if self.get_vaccine_type() is not None and \
               self.get_vaccination_date() is not None:
                self._type = "vaccine"
            if self.get_date_of_collection() is not None and \
               self.get_test_type() is not None:
                self._type = "test"
            if self.get_validity_from() is not None and \
               self.get_validity_until() is not None:
                self._type = "recovery"

        return self._type

    def set_verified(self, val):
        self.verified = val

    def get_verified(self):
        return self.verified

    def is_negative(self):
        if self.get_type() != "test":
            return True

        tr = self.get_test_result()
        if tr is None:
            return False
        return tr.is_negative()

    def add_info(self, _type, key, val):
        # Filter none value
        if val is None:
            return
        process_function = self.filter_table.get(key, None)
        # Filter value used to verify the validity of the
        # certificate
        if process_function is not None:
            process_function(val)
        else:
            self.set_info(_type, key, val)


def _iscompletelybase45compatible(data):
    s = data.decode()
    return all(c in base45.BASE45_CHARSET for c in s)


def _getnonbase45chars(data):
    s = data.decode()
    out = set()
    for c in s:
        if c not in base45.BASE45_CHARSET:
            out.add(c)
    return out


# Parse a green pass file
class GreenPassParser(object):
    def __init__(self, certification, km=GreenPassKeyManager()):
        """Parse the certificate and create a Certificate class."""
        k = km.get_default()
        self.k = k

        unstripped_data = b":".join(certification.split(b":")[1::])
        # Remove the initial HC1: part and decode the certificate
        data = b":".join(certification.strip().split(b":")[1::])

        self.not_completely_base45 = not _iscompletelybase45compatible(
            unstripped_data
        )
        self.not_base45_chars = _getnonbase45chars(unstripped_data)

        decoded = base45.b45decode(data)
        uncompressed = zlib.decompress(decoded)

        # Get the COSE message
        self.cose = CoseMessage.decode(uncompressed)

        # Extract kid and payload
        self.kid = self.get_kid_from_cose(self.cose.phdr, self.cose.uhdr)
        self.payload = cbor2.loads(self.cose.payload)

        self.qr_info = {
            k.get_release_country()[0]: self.payload[
                k.get_release_country()[0]
            ],

            k.get_release_date()[0]:    int(self.payload[
                k.get_release_date()[0]
            ]),

            k.get_expiration_date()[0]: int(self.payload[
                k.get_expiration_date()[0]
            ]),
        }

        personal_data = self.payload[k.get_personal_data()[0]][
            k.get_personal_info()[0]
        ]
        self.personal_info = {
            k.get_version()[0]:       personal_data[k.get_version()[0]],
            k.get_date_of_birth()[0]: personal_data[k.get_date_of_birth()[0]],
            k.get_first_name()[0]:    personal_data[k.get_name()[0]][
                k.get_first_name()[0]
            ],
            k.get_last_name()[0]:     personal_data[k.get_name()[0]][
                k.get_last_name()[0]
            ],
        }

        self.certificate_info = []
        if personal_data.get(k.get_vaccine()[0], None) is not None:
            # Vaccine
            self.certificate_type = k.get_vaccine()[0]
        elif personal_data.get(k.get_test()[0], None) is not None:
            # Test
            self.certificate_type = k.get_test()[0]
        elif personal_data.get(k.get_recovery()[0], None) is not None:
            # Recovery
            self.certificate_type = k.get_recovery()[0]
        else:
            raise UnrecognizedException("Unrecognized certificate type")

        for el in personal_data[self.certificate_type]:
            cert = {
                # Common
                k.get_target_disease()[0]:      el[
                    k.get_target_disease()[0]
                ],

                k.get_vaccination_country()[0]: el[
                    k.get_vaccination_country()[0]
                ],

                k.get_certificate_issuer()[0]:  el[
                    k.get_certificate_issuer()[0]
                ],

                k.get_certificate_id()[0]:      el[
                    k.get_certificate_id()[0]
                ],

                # Recovery
                k.get_first_positive_test()[0]: el.get(
                    k.get_first_positive_test()[0], None
                ),

                k.get_validity_from()[0]:       el.get(
                    k.get_validity_from()[0], None
                ),

                k.get_validity_until()[0]:      el.get(
                    k.get_validity_until()[0], None
                ),

                # Common for Test and Vaccine
                k.get_manufacturer()[0]:        el.get(
                    k.get_manufacturer()[0], None
                ),

                # Test
                k.get_test_type()[0]:           el.get(
                    k.get_test_type()[0], None
                ),

                k.get_test_name()[0]:           el.get(
                    k.get_test_name()[0], None
                ),

                k.get_date_of_collection()[0]:  el.get(
                    k.get_date_of_collection()[0], None
                ),

                k.get_test_result()[0]:         el.get(
                    k.get_test_result()[0], None
                ),

                k.get_testing_center()[0]:      el.get(
                    k.get_testing_center()[0], None
                ),

                # Vaccine
                k.get_dose_number()[0]:         int(el.get(
                    k.get_dose_number()[0], 0
                )),

                k.get_total_doses()[0]:         int(el.get(
                    k.get_total_doses()[0], 0
                )),

                k.get_vaccine_pn()[0]:          el.get(
                    k.get_vaccine_pn()[0], None
                ),

                k.get_vaccine_type()[0]:        el.get(
                    k.get_vaccine_type()[0], None
                ),

                k.get_vaccination_date()[0]:    el.get(
                    k.get_vaccination_date()[0], None
                )

            }
            self.certificate_info.append(cert)

    # Return the signature from COSE object
    def get_sign_from_cose(self):
        return self.cose.signature

    # Return the headers from COSE object
    def get_headers_from_cose(self):
        return self.cose.phdr, self.cose.uhdr

    # Isolate KID from COSE object
    @staticmethod
    def get_kid_from_cose(phdr, uhdr):
        for k in phdr.keys():
            if (isinstance(KID(), k)):
                return phdr[k]
        for k in uhdr.keys():
            if (isinstance(KID(), k)):
                return uhdr[k]
        print("Could not find KID", file=sys.stderr)
        return None

    # Get Key ID from the QRCode
    def get_kid(self):
        return self.kid

    def get_sign_alg(self):
        alg = None
        for k in self.cose.phdr.keys():
            if (isinstance(Algorithm(), k)):
                alg = self.cose.phdr[k]
        for k in self.cose.uhdr.keys():
            if (isinstance(Algorithm(), k)):
                alg = self.coseuhdr[k]

        if alg is None:
            print("Could not find Algorithm", file=sys.stderr)
            return None

        return alg.fullname

    # Set the decryption key
    def set_key(self, key):
        self.cose.key = key

    # Verify the code
    def verify(self):
        return self.cose.verify_signature()

    # Dump the content of the payload in JSON format
    def dump(self, om):
        om.rawdump(json.dumps(self.payload))

    def get_certificate(self):
        c = Certificate(self.k)
        # Invalid GP, more than one certificate info
        if len(self.certificate_info) > 1:
            return None

        for qr_info in self.qr_info.items():
            c.add_info("qr", qr_info[0], qr_info[1])
        for personal_info in self.personal_info.items():
            c.add_info("personal", personal_info[0], personal_info[1])
        for cert_info in self.certificate_info[0].items():
            c.add_info("cert", cert_info[0], cert_info[1])

        c.set_sign_alg(self.get_sign_alg())
        c.set_kid(self.get_kid())

        c.set_parent(self)

        return c


# Logic Manager, retrieve information from the certificate and set
#  output.
class LogicManager(object):
    def __init__(self, cachedir):
        """Verification of the certificate."""
        self.cachedir = cachedir

    # Verify certificate
    @staticmethod
    def verify_certificate(cert, sm, cup,
                           enable_blocklist=True,
                           raw=False,
                           verificator_key=None):

        expired = False
        blocklisted = False
        hours_to_valid = None
        remaining_hours = None

        dn = cert.get_dose_number()
        sd = cert.get_total_doses()
        positive = not cert.is_negative()
        recovery_from = cert.get_validity_from()
        recovery_until = cert.get_validity_until()
        vaccine = cert.get_vaccine_pn()
        testtype = cert.get_test_type()
        vaccinedate = cert.get_vaccination_date()
        testcollectiondate = cert.get_date_of_collection()
        cert_id = cert.get_certificate_id()

        certificate_type = cert.get_type()

        # Check test validity
        if certificate_type == "test":
            ttype = TestType(testtype)
            hours_to_valid, remaining_hours = sm.get_test_remaining_time(
                testcollectiondate, ttype.get_type()
            )

        # Check vaccine validity
        if certificate_type == "vaccine":
            hours_to_valid, remaining_hours = sm.get_vaccine_remaining_time(
                vaccinedate, vaccine, dn == sd
            )

        # Check recovery validity
        if certificate_type == "recovery":
            # TODO: implement pv recovery when available
            rtype = "default"
            hours_to_valid, remaining_hours = sm.get_recovery_remaining_time(
                rtype, recovery_from, recovery_until
            )

        if hours_to_valid is not None and remaining_hours is not None:
            if hours_to_valid < 0:
                expired = True
            elif remaining_hours <= 0:
                expired = True
            else:
                expired = False

        blocklisted = enable_blocklist and sm.check_uvci_blocklisted(cert_id)

        cert.set_expired(expired)
        cert.set_hours_to_valid(hours_to_valid)
        cert.set_remaining_hours(remaining_hours)

        # Check blocklist by ID
        cert.set_blocklisted(blocklisted)

        alg = cert.get_sign_alg()
        key = cup.get_key_coseobj(cert.get_kid(), alg=alg)
        cert.set_key(key)
        verified = cert.verify()

        unknown_cert = not cert.get_type() == "vaccine"
        unknown_cert = unknown_cert and not cert.get_type() == "test"
        unknown_cert = unknown_cert and not cert.get_type() == "recovery"

        cert.set_verified(verified)

        valid = verified
        valid = valid and not expired
        valid = valid and not positive
        valid = valid and not unknown_cert
        valid = valid and not blocklisted
        return valid



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import json
import cbor2
import requests

TESTS_URL = (
    "https://covid-19-diagnostics.jrc.ec.europa.eu/devices/export"
)


class Localized_GreenPassKeyManager(object):
    def __init__(self):
        """Manage the translation between qrcode keys and readable names."""
        pass

    def get_cert_type_long_name(self, t):
        if t == self.get_vaccine()[0]:
            return self.get_vaccine()[1]
        if t == self.get_test()[0]:
            return self.get_test()[1]
        if t == self.get_recovery()[0]:
            return self.get_recovery()[1]

    def translate_type(self, t):
        if t == "vaccine":
            return self.get_vaccine()[1]
        elif t == "test":
            return self.get_test()[1]
        elif t == "recovery":
            return self.get_recovery()[1]

        return t

    # The first value of these tuples is the key in the qrcode,
    # the second is the localized info
    def get_key(self, k):
        return k, self._data.get(k, k)

    def get_release_country(self):
        return self.get_key(1)

    def get_release_date(self):
        return self.get_key(6)

    def get_expiration_date(self):
        return self.get_key(4)

    def get_version(self):
        return self.get_key("ver")

    def get_date_of_birth(self):
        return self.get_key("dob")

    def get_name(self):
        return self.get_key("nam")

    def get_personal_data(self):
        return self.get_key(-260)

    def get_personal_info(self):
        # Personal info has a special key which clashes with the
        # release country one.
        return 1, self.get_key("_per")[1]

    def get_first_name(self):
        return self.get_key("gn")

    def get_last_name(self):
        return self.get_key("fn")

    def get_vaccine(self):
        return self.get_key("v")

    def get_test(self):
        return self.get_key("t")

    def get_recovery(self):
        return self.get_key("r")

    def get_target_disease(self):
        return self.get_key("tg")

    def get_vaccination_country(self):
        return self.get_key("co")

    def get_certificate_issuer(self):
        return self.get_key("is")

    def get_certificate_id(self):
        return self.get_key("ci")

    def get_first_positive_test(self):
        return self.get_key("fr")

    def get_validity_from(self):
        return self.get_key("df")

    def get_validity_until(self):
        return self.get_key("du")

    def get_manufacturer(self):
        return self.get_key("ma")

    def get_test_type(self):
        return self.get_key("tt")

    def get_test_name(self):
        return self.get_key("tn")

    def get_date_of_collection(self):
        return self.get_key("sc")

    def get_blocklisted(self):
        return self.get_key("_block")

    def get_test_result(self):
        return self.get_key("tr")

    def get_testing_center(self):
        return self.get_key("tc")

    def get_dose_number(self):
        return self.get_key("dn")

    def get_total_doses(self):
        return self.get_key("sd")

    def get_vaccine_pn(self):
        return self.get_key("mp")

    def get_vaccine_type(self):
        return self.get_key("vp")

    def get_vaccination_date(self):
        return self.get_key("dt")

    def get_certificate_type(self):
        return self.get_key("_typ")

    def get_verified(self):
        return self.get_key("_ver")

    def get_doses(self):
        return self.get_key("_dos")

    def get_expired_since(self):
        return self.get_key("_expired_since")

    def get_hours(self):
        return self.get_key("_hours")

    def get_hours_left(self):
        return self.get_key("_hours_left")

    def get_not_yet_valid(self):
        return self.get_key("_not_yet_valid")

    def get_hours_to_validity(self):
        return self.get_key("_hours_to_validity")

    def get_months(self):
        return self.get_key("_months")

    def get_days(self):
        return self.get_key("_days")


class IT_GreenPassKeyManager(Localized_GreenPassKeyManager):
    def __init__(self):
        """Italian localized greenpass strings."""
        self._data = {
            1: "Paese di rilascio",
            6: "Data di rilascio",
            4: "Data di scadenza",
            "ver": "Versione",
            "dob": "Data di Nascita",
            "nam": "Nome",
            -260: "Dati personali",
            "_per": "Informazioni personali",
            "gn": "Nome",
            "fn": "Cognome",
            "v": "Vaccino",
            "t": "Test",
            "r": "Recupero",
            "tg": "Malattia",
            "co": "Paese di vaccinazione o test",
            "is": "Ente di rilascio",
            "ci": "ID Certificato",
            "fr": "Data primo test positivo",
            "df": "Valido da",
            "du": "Valido fino a",
            "ma": "Produttore e tipo",
            "tt": "Tipo di test",
            "tn": "Nome del test",
            "sc": "Data di analisi",
            "dt": "Data di vaccinazione",
            "tr": "Risultato del test",
            "tc": "Centro di test",
            "dn": "Dose numero",
            "sd": "Dosi totali",
            "mp": "Codice del vaccino",
            "vp": "Tipo di vaccino",
            "_typ": "Tipo di certificato",
            "_ver": "Verificato",
            "_dos": "Dosi",
            "_block": "Bloccato",
            "_False": "No",
            "_True": "Sì",
            "_expired_since": "Scaduto da",
            "_hours": "Ore",
            "_hours_left": "Ore rimanenti",
            "_not_yet_valid": "Non ancora valido",
            "_hours_to_validity": "Ore rimanenti per risultare valido",
            "_days": "Giorni",
            "_months": "Mesi"
        }

    @staticmethod
    def get_date_format():
        return "Data {}"


class DE_GreenPassKeyManager(Localized_GreenPassKeyManager):
    def __init__(self):
        """German localized greenpass strings."""
        self._data = {
            1: "Freigabeland",
            6: "Veröffentlichungsdatum",
            4: "Verfallsdatum",
            "ver": "Version",
            "dob": "Geburtsdatum",
            "nam": "Name",
            -260: "Persönliche Daten",
            "_per": "Persönliche Informationen",
            "gn": "Vorname",
            "fn": "Name",
            "v": "Impfung",
            "t": "Prüfung",
            "r": "Erholung",
            "tg": "Zielkrankheit",
            "co": "Impfland",
            "is": "Zertifikataussteller",
            "ci": "Zertifikats-ID",
            "fr": "Erster Positiver Test",
            "df": "Gültigkeit ab",
            "du": "Gültigkeit bis",
            "ma": "Hersteller und Typ",
            "tt": "Testtyp",
            "tn": "Testname",
            "sc": "Datum der Abholung",
            "tr": "Testergebnis",
            "tc": "Testzentrum",
            "dn": "Dosisnummer",
            "sd": "Gesamtdosen",
            "mp": "Impfcode",
            "vp": "Impftyp",
            "dt": "Impfdatum",
            "_typ": "Art des Zertifikats",
            "_ver": "Verifiziert",
            "_dos": "Dosen",
            "_block": "Verstopft",
            "_False": "Nein",
            "_True": "Ja",
            "_expired_since": "Expired Since",
            "_hours": "Stunden",
            "_hours_left": "Stunden übrig",
            "_not_yet_valid": "noch nicht gültig",
            "_hours_to_validity": "Stunden bis Gültigkeit",
            "_days": "Tage",
            "_months": "Monate"
        }

    @staticmethod
    def get_date_format():
        return "{}datum"


class EN_GreenPassKeyManager(Localized_GreenPassKeyManager):
    def __init__(self):
        """English localized greenpass strings."""
        self._data = {
            1: "Release Country",
            6: "Release Date",
            4: "Expiration Date",
            "ver": "Version",
            "dob": "Date of Birth",
            "nam": "Name",
            -260: "Personal Data",
            "_per": "Personal Info",
            "gn": "First Name",
            "fn": "Family Name",
            "v": "Vaccine",
            "t": "Test",
            "r": "Recovery",
            "tg": "Target Disease",
            "co": "Vaccination or Test Country",
            "is": "Certificate Issuer",
            "ci": "Certificate ID",
            "fr": "First Positive Test",
            "df": "Validity From",
            "du": "Validity Until",
            "ma": "Manufacturer and Type",
            "tt": "Test type",
            "tn": "Test name",
            "sc": "Date of collection",
            "tr": "Test result",
            "tc": "Testing center",
            "dn": "Dose Number",
            "sd": "Total Doses",
            "mp": "Vaccine Product Number",
            "vp": "Vaccine Type",
            "dt": "Vaccination Date",
            "_typ": "Certificate Type",
            "_ver": "Verified",
            "_dos": "Doses",
            "_block": "Blocklisted",
            "_False": "False",
            "_True": "True",
            "_expired_since": "Expired Since",
            "_hours": "Hours",
            "_hours_left": "Hours left",
            "_not_yet_valid": "Not yet valid",
            "_hours_to_validity": "Hours to validity",
            "_days": "Days",
            "_months": "Months"
        }

    @staticmethod
    def get_date_format():
        return "{} Date"


class GreenPassKeyManager(object):
    def __init__(self):
        """Dispatch the localized key values."""

    def get_localization(self, language="en"):
        if language == "it":
            return IT_GreenPassKeyManager()
        elif language == "de":
            return DE_GreenPassKeyManager()
        else:
            return EN_GreenPassKeyManager()
        return self.get_default()

    def get_default(self):
        return self.get_localization(language="en")


# Vaccine names
class Vaccine(object):
    def __init__(self, t):
        """Translate vaccine code to human-readable name."""
        self.t = t
        self.pretty_name = {
            "EU/1/20/1507": "Moderna",
            "EU/1/20/1525": "Janssen",
            "EU/1/20/1528": "Pfizer",
            "EU/1/21/1529": "AstraZeneca",
            "EU/1/XX/XXX1": "Sputnik-V",
            "EU/1/XX/XXX2": "CVnCoV",
            "EU/1/XX/XXX3": "EpiVacCorona",
            "EU/1/XX/XXX4": "BBIBP-CorV",
            "EU/1/XX/XXX5": "CoronaVac",
        }

    def get_pretty_name(self):
        return self.pretty_name.get(self.t, self.t)


# Manufacturer names
class Manufacturer(object):
    def __init__(self, t, cachedir=''):
        """Translate manufacturer code to human-readable name."""
        self.t = t
        # Vaccines
        self.pretty_name = {
            "ORG-100001699": "AstraZeneca",
            "ORG-100030215": "Biontech",
            "ORG-100001417": "Janssen",
            "ORG-100031184": "Moderna",
            "ORG-100006270": "Curevac",
            "ORG-100013793": "CanSino",
            "ORG-100020693": "Sinopharm",
            "ORG-100010771": "Sinopharm",
            "ORG-100024420": "Sinopharm",
            "ORG-100032020": "Novavax"
        }
        # Testes
        if cachedir == '':
            self.pretty_name.update(self.get_tests_pn())
        else:
            self.pretty_name.update(self.get_cached_tests_pn(cachedir))

    @staticmethod
    def get_tests_pn():
        o = {}
        try:
            r = requests.get(TESTS_URL, allow_redirects=True, timeout=10)
        except requests.exceptions.ReadTimeout:
            # The operation timed out, return empty value
            return o

        if r.status_code != 200:
            # The operation did not succeded, return empty value
            return o

        try:
            tests = json.loads(r.text)
            for el in tests["deviceList"]:
                o[el["id_device"]] = el["commercial_name"]
        # TODO: Be more specific on the exceptions
        except Exception:
            # The data cannot be downloaded, return value correctly parsed
            print("Warning: cannot download test data", file=sys.stderr)
            return o

        return o

    def get_cached_tests_pn(self, cachedir):
        testcache = os.path.join(cachedir, "tests")

        os.makedirs(cachedir, exist_ok=True)

        if not os.path.exists(testcache):
            with open(testcache, 'wb') as f:
                cbor2.dump(self.get_tests_pn(), f)

        with open(testcache, 'rb') as f:
            try:
                tests = cbor2.load(f)
            except Exception:
                print("Corrupted cache {} removing".format(testcache),
                      file=sys.stderr)
                os.remove(testcache)
                tests = {}

        return tests

    def get_pretty_name(self):
        return self.pretty_name.get(self.t, self.t)


# Disease names
class Disease(object):
    def __init__(self, t):
        """Translate disease code to human-readable name."""
        self.t = t
        self.pretty_name = {
            "840539006": "Covid19"
        }

    def get_pretty_name(self):
        return self.pretty_name.get(self.t, self.t)


class TestType(object):
    def __init__(self, t):
        """Translate test code to human-readable name."""
        self.t = t
        self._type = {
            "LP6464-4":   "molecular",
            "LP217198-3": "rapid"
        }
        self.pretty_name = self._type

    def get_type(self):
        return self._type.get(self.t, self.t)

    def get_pretty_name(self):
        return self.pretty_name.get(self.t, self.t)



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import os
import sys
import cbor2
import pytz
import json
import requests
from datetime import datetime
from tzlocal import get_localzone

from greenpass.URLs import BASE_URL_DGC


# Retrieve settings from unified API endpoint
class SettingsManager(object):
    def __init__(self, cachedir='', consider_recovery_expiration=False):
        """Download and parse the settings from API endpoint."""
        self.at_date = None
        self.cachedir = cachedir
        self.blocklist = set()
        if cachedir != '':
            self.get_cached_settings()
        else:
            self.get_settings()
        self.consider_recovery_expiration = consider_recovery_expiration

    def get_cached_settings(self):
        settings = os.path.join(self.cachedir, "settings")

        os.makedirs(self.cachedir, exist_ok=True)

        if not os.path.exists(settings):
            with open(settings, 'wb') as f:
                cbor2.dump(self.get_settings(), f)

        with open(settings, 'rb') as f:
            (self.vaccines,
             self.recovery,
             self.test,
             self.blocklist) = cbor2.load(f)

    # Dispatchers
    def dispatch_vaccine(self, setting):
        vaccine_template = {
            "complete": {
                "start_day": -1,
                "end_day": -1
            },
            "not_complete": {
                "start_day": -1,
                "end_day": -1
            }
        }
        field_name = setting["name"]
        field_type = setting["type"]
        field_value = setting["value"]
        vtype_regex = r"((?:not_)?complete)"
        daytype_regex = r"((?:start|end)_day)"
        if self.vaccines.get(field_type, None) is None:
            self.vaccines[field_type] = vaccine_template

        vtype_re = re.search(vtype_regex, field_name)
        daytype_re = re.search(daytype_regex, field_name)

        vtype = vtype_re.group(1)
        daytype = daytype_re.group(1)

        self.vaccines[field_type][vtype][daytype] = int(field_value)

    def dispatch_blocklist(self, setting):
        self.blocklist = self.blocklist.union(set(
            setting["value"].split(";")[:-1])
        )

    def dispatch_operating_system(self, setting):
        # Ignore app specific options
        pass

    def dispatch_test(self, setting):
        field_name = setting["name"]
        field_value = setting["value"]
        if "molecular" in field_name:
            ttype = "molecular"
        elif "rapid" in field_name:
            ttype = "rapid"

        if "start_hours" in field_name:
            hourtype = "start_hours"
        elif "end_hours" in field_name:
            hourtype = "end_hours"

        self.test[ttype][hourtype] = int(field_value)

    def dispatch_recovery(self, setting):
        field_name = setting["name"]
        field_value = setting["value"]

        # Normal recovery
        rtarget = self.recovery["default"]
        # Recovery after vaccine
        if "recovery_pv_" in field_name:
            rtarget = self.recovery["pv"]

        if "start_day" in field_name:
            rtarget["start_day"] = int(field_value)
        elif "end_day" in field_name:
            rtarget["end_day"] = int(field_value)

    def dispatch_setting(self, setting):
        field_name = setting["name"]
        if "vaccine" in field_name:
            self.dispatch_vaccine(setting)
        elif "recovery" in field_name:
            self.dispatch_recovery(setting)
        elif "test" in field_name:
            self.dispatch_test(setting)
        elif "ios" == field_name or "android" == field_name:
            self.dispatch_operating_system(setting)
        elif "black_list_uvci" == field_name:
            self.dispatch_blocklist(setting)
        else:
            print("[~] Unknown field {}".format(setting["name"]))

    def get_settings(self):
        r = requests.get("{}/settings".format(BASE_URL_DGC))
        if r.status_code != 200:
            print("[-] Error from API")
            sys.exit(1)

        self.vaccines = {}
        self.recovery = {
            "default": {},
            "pv": {}
        }
        self.test = {
            "molecular": {},
            "rapid": {}
        }

        settings = json.loads(r.text)
        # Dispatch and create the dicts
        for setting in settings:
            self.dispatch_setting(setting)

        return self.vaccines, self.recovery, self.test, self.blocklist

    # Return the time that a test is still valid, negative time if expired
    def get_test_remaining_time(self, test_date, ttype):
        hours = self.test.get(ttype, 0)

        try:
            seconds_since_test = (self.checktime() - test_date).total_seconds()
            hours_since_test = seconds_since_test / (60 * 60)
        except Exception as e:
            print(e, file=sys.stderr)
            return 0, 0

        valid_start = (hours_since_test - hours["start_hours"])
        valid_end = (hours["end_hours"] - hours_since_test)

        return valid_start, valid_end

    # Return the time that a vaccine is still valid, negative
    # time if expired
    def get_vaccine_remaining_time(self, vaccination_date, vtype, full):
        if full:
            selector = "complete"
        else:
            selector = "not_complete"

        days = self.vaccines.get(
            vtype, {"complete": 0, "not_complete": 0}
        )[selector]

        try:
            seconds_since_vaccine = (
                self.checktime() - vaccination_date
            ).total_seconds()
            hours_since_vaccine = seconds_since_vaccine / (60 * 60)
        except Exception as e:
            print(e, file=sys.stderr)
            return 0, 0

        valid_start = (hours_since_vaccine - days["start_day"] * 24)
        valid_end = (days["end_day"] * 24 - hours_since_vaccine)

        return int(valid_start), int(valid_end)

    # Return the time that a recovery certification is still valid, negative
    # time if expired
    def get_recovery_remaining_time(self, rtype,
                                    recovery_from, recovery_until):
        days = self.recovery[rtype]

        try:
            seconds_since_recovery = (
                self.checktime() - recovery_from
            ).total_seconds()
            hours_since_recovery = seconds_since_recovery / (60 * 60)
        except Exception as e:
            print(e, file=sys.stderr)
            return 0, 0

        valid_start = (hours_since_recovery - days["start_day"] * 24)
        valid_end = (days["end_day"] * 24 - hours_since_recovery)

        valid_until = (recovery_until - self.checktime()).total_seconds()
        valid_until = valid_until / (60 * 60)

        # The recovery contains a validity until indication, this seems
        # to be ignored by verification applications, the following
        # code re-enables the expiration date contained in the
        # certificate.
        if self.consider_recovery_expiration:
            valid_end = min(valid_end, valid_until)

        return int(valid_start), int(valid_end)

    def get_blocklist(self):
        return self.blocklist

    def check_uvci_blocklisted(self, uvci):
        return uvci in self.blocklist

    def checktime(self):
        if self.at_date is None:
            d = datetime.now(pytz.utc)
        else:
            d = self.at_date
        return d

    def set_at_date(self, at_date):
        tz = get_localzone()
        # Sanity check to see which are the value inputted
        # Date time, seconds and timezone
        if re.match(r"\d{4}-\d{2}-\d{2}-\d{2}:\d{2}:\d{2}[+-][:0-9]{2,4}",
                    at_date):
            self.at_date = datetime.strptime(at_date, "%Y-%m-%d-%H:%M:%S%z")
        # Date, time and timezone
        elif re.match(r"\d{4}-\d{2}-\d{2}-\d{2}:\d{2}[+-][:0-9]{2,4}",
                      at_date):
            self.at_date = datetime.strptime(at_date, "%Y-%m-%d-%H:%M%z")
        # Date time and seconds
        elif re.match(r"\d{4}-\d{2}-\d{2}-\d{2}:\d{2}:\d{2}", at_date):
            self.at_date = datetime.strptime(at_date, "%Y-%m-%d-%H:%M:%S")
            self.at_date = self.at_date.replace(tzinfo=tz)
        # Date and time
        elif re.match(r"\d{4}-\d{2}-\d{2}-\d{2}:\d{2}", at_date):
            self.at_date = datetime.strptime(at_date, "%Y-%m-%d-%H:%M")
            self.at_date = self.at_date.replace(tzinfo=tz)
        # Only the date
        elif re.match(r"\d{4}-\d{2}-\d{2}", at_date):
            self.at_date = datetime.strptime(at_date, "%Y-%m-%d")
            self.at_date = self.at_date.replace(tzinfo=tz)
        else:
            print("[-] Unrecognized time format {}".format(at_date))
            sys.exit(1)



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
import sys
import fitz
from PIL import Image
from pyzbar import pyzbar


# Class to get input data from various sources.
# Current supported:
# TXT Data
# PNG Image
# PDF Documents
class InputTransformer(object):
    def __init__(self, path, filetype):
        """Transform input to the format understood by the application."""
        if filetype == "txt":
            if path == "-":
                outdata = bytes(
                    sys.stdin.read().split("\n")[0].encode("ASCII")
                )
            else:
                with open(path, 'rb') as f:
                    outdata = f.read()
        else:
            if filetype == "png":
                img = Image.open(path)
            elif filetype == "pdf":
                # Convert PDF to JPG
                pdf_file = fitz.open(path)
                imagebytes = pdf_file.extract_image(6)["image"]
                img = Image.open(io.BytesIO(imagebytes))
            else:
                print("[-] file format {} not recognized".format(filetype),
                      file=sys.stderr)

            decoded = pyzbar.decode(img)
            if len(decoded) < 1:
                print("[-] Value not found", file=sys.stderr)
                sys.exit(1)
            output = decoded[0]
            if output.type != "QRCODE":
                print("[-] Not a qrcode", file=sys.stderr)
                sys.exit(1)

            outdata = output.data
        self.data = outdata

    def get_data(self):
        return self.data



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
from termcolor import colored

from greenpass.data import Disease
from greenpass.data import Manufacturer
from greenpass.data import Vaccine
from greenpass.data import TestType
from greenpass.data import GreenPassKeyManager


class NoneOutput(object):
    def __init__(self, colored=colored):
        """Null output, /dev/null sink."""
        pass

    def add_general_info(self, infoname, infoval):
        pass

    def add_cert_info(self, infoname, infoval):
        pass

    def add_general_info_ok(self, infoname, infoval):
        pass

    def add_general_info_info(self, infoname, infoval):
        pass

    def add_general_info_warning(self, infoname, infoval):
        pass

    def add_general_info_error(self, infoname, infoval):
        pass

    def add_cert_info_ok(self, infoname, infoval):
        pass

    def add_cert_info_info(self, infoname, infoval):
        pass

    def add_cert_info_warning(self, infoname, infoval):
        pass

    def add_cert_info_error(self, infoname, infoval):
        pass

    def add_remaining_time(self, certtype, certdate, level,
                           remaining_days, km):
        pass

    def get_not_yet_valid(self, hours_to_valid, km):
        pass

    def get_expired(self, remaining_hours, km):
        pass

    def get_hours_left(self, remaining_hours, km):
        pass

    def get_months_left(self, remaining_hours, km):
        pass

    def dump(self, file=sys.stdout):
        pass

    def dump_settings(self, sm):
        pass

    def dump_cose(self, phdr, uhdr, signature):
        pass

    def rawdump(self, data):
        pass

    def print_cert(self, cert, kmg=GreenPassKeyManager(),
                   cachedir='', language=''):
        pass


# Output Manager, manages output and dumps to files and stdout
class OutputManager(NoneOutput):

    def __init__(self, colored=colored):
        """Default printer manager."""
        self.out = ""
        self.colored = colored

    def add_general_info(self, infoname, infoval):
        if infoval is not None:
            self.out += "{:30s} {}\n".format(infoname, infoval)

    def add_cert_info(self, infoname, infoval):
        if infoval is not None:
            self.out += "  {:28s} {}\n".format(infoname, infoval)

    def add_general_info_ok(self, infoname, infoval):
        if infoval is not None:
            self.add_general_info(infoname, self.colored(infoval, "green"))

    def add_general_info_info(self, infoname, infoval):
        if infoval is not None:
            self.add_general_info(infoname, self.colored(infoval, "blue"))

    def add_general_info_warning(self, infoname, infoval):
        if infoval is not None:
            self.add_general_info(infoname, self.colored(infoval, "yellow"))

    def add_general_info_error(self, infoname, infoval):
        if infoval is not None:
            self.add_general_info(infoname, self.colored(infoval, "red"))

    def add_cert_info_ok(self, infoname, infoval):
        if infoval is not None:
            self.add_cert_info(infoname, self.colored(infoval, "green"))

    def add_cert_info_info(self, infoname, infoval):
        if infoval is not None:
            self.add_cert_info(infoname, self.colored(infoval, "blue"))

    def add_cert_info_warning(self, infoname, infoval):
        if infoval is not None:
            self.add_cert_info(infoname, self.colored(infoval, "yellow"))

    def add_cert_info_error(self, infoname, infoval):
        if infoval is not None:
            self.add_cert_info(infoname, self.colored(infoval, "red"))

    def add_remaining_time(self, certtype, certdate, level,
                           remaining_days, km):
        if level == 0:
            color = "white"
        if level == 1:
            color = "green"
        if level == 2:
            color = "yellow"
        if level == 3:
            color = "red"

        self.out += "  {:28s} {} ({})\n".format(
            km.get_date_format().format(certtype),
            self.colored(certdate, color),
            self.colored(remaining_days, color)
        )

    def get_not_yet_valid(self, hours_to_valid, km):
        return "{}, {:.0f} {}, {} {}".format(
                km.get_not_yet_valid()[1],
                hours_to_valid,
                km.get_hours_to_validity()[1],
                int(hours_to_valid / 24),
                km.get_days()[1]
        )

    def get_expired(self, remaining_hours, km):
        return "{} {:.0f} {}, {} {}".format(
                    km.get_expired_since()[1],
                    -remaining_hours,
                    km.get_hours()[1],
                    -int(remaining_hours / 24),
                    km.get_days()[1]
                )

    def get_hours_left(self, remaining_hours, km):
        return "{:.0f} {} ({} {})".format(
                    remaining_hours,
                    km.get_hours_left()[1],
                    int(remaining_hours / 24),
                    km.get_days()[1]
                )

    def get_months_left(self, remaining_hours, km):
        return "{:.0f} {}, {} {}, ~ {} {}".format(
                remaining_hours,
                km.get_hours_left()[1],
                int(remaining_hours / 24),
                km.get_days()[1],
                round(remaining_hours / 24 / 30),
                km.get_months()[1]
        )

    def dump(self, file=sys.stdout):
        print(self.out, file=file)

    def dump_settings(self, sm):
        print("Tests")
        for el in sm.test.items():
            print("  {} not before: {:4d} hours not after: {:4d} hours".format(
                self.colored("{:25s}".format(el[0]), "blue"),
                el[1]["start_hours"],
                el[1]["end_hours"])
            )

        print("\nCertifications")
        for rtype in ( "default", "pv" ):
            rstring = "recovery"
            if rtype == "pv":
                rstring = "recovery+vaccine"

            print("  {} not before: {:4d} days  not after: {:4d} days".format(
                self.colored("{:25s}".format(rstring), "blue"),
                sm.recovery[rtype]["start_day"],
                sm.recovery[rtype]["end_day"])
            )

        print("\nVaccines")
        for vac in sm.vaccines.items():
            for el in vac[1].items():
                print((
                    "  {} {} not before: {:4d} days  "
                    "not after: {:4d} days"
                ).format(
                    self.colored("{:12s}".format(el[0]), "blue"),
                    self.colored("{:20s}".format(
                        Vaccine(vac[0]).get_pretty_name()
                    ), "yellow"),
                    el[1]["start_day"], el[1]["end_day"]
                ))

        print("\nBlocked Pass IDs")
        for gp in sm.blocklist:
            print("  {}".format(gp))
        print()

    def dump_cose(self, phdr, uhdr, signature):
        print(self.colored("PHDR:", "green"), phdr)
        print(self.colored("UHDR:", "red"),   uhdr)
        print(self.colored("Sign:", "cyan"),  signature)

    def rawdump(self, data):
        print(data)

    def _print_common_fields(self, cert, km, cachedir):
        self.add_general_info_info(
            km.get_certificate_type()[1],
            km.translate_type(cert.get_type())
        )

        for k, v in cert.get_info()["qr"].items():
            self.add_general_info(km.get_key(k)[1], v)

        self.add_general_info(
            km.get_release_date()[1], cert.get_release_date()
        )

        self.add_general_info(
            km.get_expiration_date()[1], cert.get_expiration_date()
        )

        for k, v in cert.get_info()["personal"].items():
            self.add_general_info(km.get_key(k)[1], km.get_key(str(v))[1])

        self.add_cert_info_info(
            km.get_target_disease()[1],
            Disease(cert.get_target_disease()).get_pretty_name()
        )
        for k, v in cert.get_info()["cert"].items():
            self.add_cert_info(km.get_key(k)[1], km.get_key(str(v))[1])

        self.add_cert_info(
            km.get_certificate_id()[1], cert.get_certificate_id()
        )

        if cert.get_blocklisted():
            printfun = self.add_cert_info_error
        else:
            printfun = self.add_cert_info_ok
        printfun(
            km.get_blocklisted()[1],
            km.get_key("_"+str(cert.get_blocklisted()))[1]
        )

    def _print_vaccine_fields(self, cert, km, cachedir):
        if cert.get_dose_number() == cert.get_total_doses():
            printfun = self.add_cert_info_ok
        else:
            printfun = self.add_cert_info_warning

        if cert.get_total_doses() > 0:
            printfun(
                km.get_doses()[1],
                f"{cert.get_dose_number()}/{cert.get_total_doses()}"
            )

        self.add_cert_info_info(
            km.get_manufacturer()[1],
            Manufacturer(cert.get_manufacturer(), cachedir).get_pretty_name()
        )

        self.add_cert_info_info(
            km.get_vaccine_pn()[1],
            Vaccine(cert.get_vaccine_pn()).get_pretty_name()
        )

        self.add_cert_info(
            km.get_vaccine_type()[1],
            Vaccine(cert.get_vaccine_type()).get_pretty_name()
        )

    def _print_test_fields(self, cert, km, cachedir):
        self.add_cert_info_info(
            km.get_test_type()[1],
            TestType(cert.get_test_type()).get_pretty_name()
        )

        test_result = cert.get_test_result()

        if test_result.is_positive():
            printfun = self.add_cert_info_error
        elif test_result.is_negative():
            printfun = self.add_cert_info_ok
        else:
            printfun = self.add_cert_info_warning
        printfun(
            km.get_test_result()[1],
            test_result
        )

    def _print_recovery_fields(self, cert, km, cachedir):
        self.add_cert_info_info(
            km.get_validity_until()[1],
            cert.get_validity_until()
        )

    def print_cert(self, cert, kmg=GreenPassKeyManager(),
                   cachedir='', language="en"):
        km = kmg.get_localization(language)

        # TODO: remove when recovery+vaccine logic is implemented
        if cert.get_type() == "recovery":
            self.add_general_info_warning(
                "",
                "\n[ WARNING ] "
                "The Recovery+Vaccine logic is currently not implemented."
                "see https://github.com/berdav/greenpass/issues/37"
            )

        # Print warning if the certificate contains non base45 chars
        if cert.get_parent().not_completely_base45:
            self.add_general_info_warning(
                "",
                "\n[ WARNING ] "
                "This certificate can generate problems with base45 decoders"
                "\n            "
                "it contains the following non base45 characters"
            )
            for c in cert.get_parent().not_base45_chars:
                self.add_general_info_warning("            "+repr(c), "")

        self._print_common_fields(cert, km, cachedir)

        hours_to_valid = cert.get_hours_to_valid()
        remaining_hours = cert.get_remaining_hours()

        if hours_to_valid < 0:
            level = 3
            remaining_days = self.get_not_yet_valid(-hours_to_valid, km)
        elif remaining_hours <= 0:
            level = 3
            remaining_days = self.get_expired(remaining_hours, km)
        elif remaining_hours * 24 < 14:
            level = 2
            remaining_hours = self.get_hours_left(remaining_hours, km)
            remaining_days = remaining_hours
        else:
            level = 1
            remaining_days = self.get_months_left(remaining_hours, km)

        date = None
        if cert.get_type() == "test":
            self._print_test_fields(cert, km, cachedir)
            date = cert.get_date_of_collection()
        elif cert.get_type() == "vaccine":
            self._print_vaccine_fields(cert, km, cachedir)
            date = cert.get_vaccination_date()
        elif cert.get_type() == "recovery":
            self._print_recovery_fields(cert, km, cachedir)
            date = cert.get_validity_from()

        self.add_remaining_time(
            km.get_key(cert.get_type()[0])[1], date,
            level, remaining_days, km
        )

        if cert.get_verified():
            printfun = self.add_general_info_ok
        else:
            printfun = self.add_general_info_error
        printfun(
            km.get_verified()[1],
            km.get_key("_"+str(cert.get_verified()))[1]
        )



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import shutil
from distutils.core import setup
import setuptools

print("Using setuptools {}".format(setuptools.__version__), file=sys.stderr)

maintool = "greenpass"

build_scripts = os.path.join("build", "scripts")

maintool_dst_path = os.path.join(build_scripts, maintool)

os.makedirs(build_scripts, exist_ok=True)
shutil.copyfile(maintool + ".py", maintool_dst_path)

with open("README.md") as f:
    description = f.read()

setup(
    name             = 'greenpass',
    scripts          = [ maintool_dst_path ],
    version          = '4.2.6',
    license          = 'LGPLv3',
    description      = 'Scriptable green pass verifier',
    long_description = description,
    long_description_content_type='text/markdown',
    packages         = [ "greenpass" ],
    author           = 'Davide Berardi',
    author_email     = 'berardi.dav@gmail.com',
    url              = 'https://github.com/berdav/greenpass',
    download_url     = 'https://github.com/berdav/greenpass/archive/refs/tags/v4.2.6.zip',
    keywords         = [ 'covid19', 'greenpass', 'certificates', 'authorization' ],
    install_requires = [
        'base45',
        'bs4',
        'PyMuPDF',
        'pytz',
        'cbor2',
        'colorama',
        'Pillow',
        'pyasn1',
        'pyOpenSSL',
        'pyzbar',
        'requests',
        'termcolor',
        'tzlocal',
        'cose'
    ],
    classifiers      = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Healthcare Industry',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ]
)



#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import json
import base64
import requests
from OpenSSL import crypto
from binascii import hexlify
from bs4 import BeautifulSoup
from cose.keys import CoseKey

from greenpass.URLs import BASE_URL_DGC, BASE_URL_NHS, BASE_URL_DGCG


# Update certificate signer
class CertificateUpdater(object):
    def __init__(self):
        """Download certificates from the remote endpoint."""
        self.verbose = False

    def set_verbose(self):
        self.verbose = True

    # Get KEY index from online status page
    @staticmethod
    def _get_kid_idx(kid, _type="dgc"):
        if _type == "dgc":
            r = requests.get("{}/signercertificate/status".format(
                BASE_URL_DGC
            ))
        elif _type == "nhs":
            r = requests.get("{}/pubkeys/keys.json".format(BASE_URL_NHS))
        else:
            return ("unk", -1)
        if r.status_code != 200:
            print("[-] Error from API")
            sys.exit(1)
        i = 0
        hexkid = hexlify(kid)
        for x in json.loads(r.text):
            if _type == "dgc":
                targetkid = hexlify(base64.b64decode(x))
                out = i
            if _type == "nhs":
                targetkid = hexlify(base64.b64decode(x["kid"]))
                out = targetkid
            if targetkid == hexkid:
                return (_type, out)
            i += 1
        return (_type, -1)

    # Dispatch to correct Key IDX retrieve function
    def get_kid_idx(self, kid):
        k = self._get_kid_idx(kid, "nhs")
        if k[1] != -1:
            return k
        k = self._get_kid_idx(kid, "dgc")
        if self.verbose:
            print("[ ] Kid: {} idx: {}".format(base64.b64encode(kid), k[1]))

        if k[1] != -1:
            return k

        print("[-] Could not find certification authority")
        sys.exit(1)

    # Get key from DGC style repository
    @staticmethod
    def get_key_dgc(idx):
        certificate = None
        r = requests.get("{}".format(BASE_URL_DGCG))
        if r.status_code != 200:
            print("[-] Error from API")
            sys.exit(1)

        soup = BeautifulSoup(r.text, 'html.parser')
        trust_list_json = soup.find("code", {"id": "trust-list-json"})
        trust_list = json.loads(trust_list_json.string)
        target = base64.b64encode(idx).decode()
        for country in trust_list["dsc_trust_list"].values():
            for el in country["keys"]:
                if el["kid"] == target:
                    certificate = base64.b64decode(el["x5c"][0])
                    break

        return certificate

    # Return public key
    def loadpubkey(self, certificate):
        # Load certificate and dump the pubkey
        x509 = crypto.load_certificate(crypto.FILETYPE_ASN1, certificate)

        if self.verbose:
            subject = ' '.join(map(
                lambda x: x[1].decode(),
                x509.get_subject().get_components()
            ))
            print("[ ] Signed with public key from")
            print("    {}".format(subject))
        pubkey = crypto.dump_publickey(crypto.FILETYPE_ASN1, x509.get_pubkey())
        return pubkey

    @staticmethod
    def extractpubkey(pubkey):
        return pubkey[26::]

    # Get key from NHS style repository
    @staticmethod
    def get_key_nhs(idx):
        r = requests.get("{}/pubkeys/keys.json".format(BASE_URL_NHS))
        for x in json.loads(r.text):
            targetkid = hexlify(base64.b64decode(x["kid"]))

            if targetkid == idx:
                return base64.b64decode(x["publicKey"])

    # Retrieve key from remote repository
    def get_certificate(self, kid):
        keytype, idx = self.get_kid_idx(kid)

        if keytype == "nhs":
            certificate = self.get_key_nhs(idx)
        elif keytype == "dgc":
            certificate = self.get_key_dgc(kid)

        return certificate

    def get_key(self, kid):
        certificate = self.get_certificate(kid)

        # Try to load the certificate
        try:
            pubkey = self.loadpubkey(certificate)
        # TODO: Be more specific on the exceptions
        except Exception:
            pubkey = certificate

        return pubkey

    # Retrieve key and convert to coseobj
    def get_key_coseobj(self, kid, alg="ES256"):
        pubkey = self.get_key(kid)
        return self.getcoseobj(pubkey, alg)

    # Return COSE object from public key
    def getcoseobj(self, pubkey, alg="ES256"):
        if self.verbose:
            print("[ ] Algorithm: {}".format(alg))

        if alg == "ES256":
            return self._get_es256_cose_obj(pubkey)
        if alg == "PS256":
            return self._get_ps256_cose_obj(pubkey)
        print("[ ] Unknown algorithm: {}".format(alg), file=sys.stderr)
        return None

    @staticmethod
    def _get_ps256_cose_obj(pubkey):
        # Get N and E from the key
        n = pubkey[32:-5]
        e = pubkey[-3::]
        kattr = {
                "KTY":    "RSA",
                "CURVE":  "P_256",
                "ALG":    "PS256",
                "E": e,
                "N": n
        }
        return CoseKey.from_dict(kattr)

    def _get_es256_cose_obj(self, pubkey):
        pubkey = self.extractpubkey(pubkey)
        # X is the first 32 bits, Y are the remaining ones
        x = pubkey[1:int(len(pubkey)/2) + 1]
        y = pubkey[int(len(pubkey)/2) + 1::]

        # Create COSE key
        kattr = {
                "KTY":   "EC2",
                "CURVE": "P_256",
                "ALG":   "ES256",
                "X":     x,
                "Y":     y
        }
        return CoseKey.from_dict(kattr)


# Cached version of Certificate Updater,
#  saves and retrieves public keys using a cache directory
class CachedCertificateUpdater(CertificateUpdater):
    def __init__(self, cachedir):
        """Download certificates from the remote endpoint and cache them."""
        self.cachedir = cachedir
        os.makedirs(cachedir, exist_ok=True)
        super(CachedCertificateUpdater, self).__init__()

    def get_certificate(self, kid):
        # Replace / with a value that cannot be found in base64
        enckid = base64.b64encode(kid).decode().replace("/", ".")
        cachepath = os.path.join(self.cachedir, enckid)
        superclass = super(CachedCertificateUpdater, self)

        if not os.path.exists(cachepath):
            with open(cachepath, "wb") as f:
                f.write(superclass.get_certificate(kid))

        with open(cachepath, "rb") as f:
            keybytes = f.read()

        return keybytes


class ForcedCertificateUpdater(CertificateUpdater):
    def __init__(self, path):
        """Force the key used to verify the certificate."""
        self.keypath = path
        super(ForcedCertificateUpdater, self).__init__()

    def get_certificate(self, _kid):
        with open(self.keypath, "rb") as f:
            keybytes = f.read()

        return keybytes
